# Copyright 2023 iwatake2222
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import time
import os
import pandas as pd

from .top_runner import TopRunner
from .utility import create_logger


logger = create_logger(__name__, log_filename='rotop.log')


class DataContainer:
  MAX_ROW_CSV = 600
  MAX_NUM_HISTORY = 100

  def __init__(self, write_csv=False):
    now = datetime.datetime.now()
    if write_csv:
      self.csv_dir_name = now.strftime('./rotop_%Y%m%d_%H%M%S')
      os.mkdir(self.csv_dir_name)
    else:
      self.csv_dir_name = None
    self.csv_index = 0
    self.df_cpu = pd.DataFrame()
    self.df_mem = pd.DataFrame()
    self.df_cpu_history = pd.DataFrame()
    self.df_mem_history = pd.DataFrame()

  def run(self, top_runner: TopRunner, lines: list[str], num_process: int):
    if top_runner.col_range_command and top_runner.col_range_command[0] > 0:
      df_cpu_current, df_mem_current = self.create_df_from_top(top_runner, lines, num_process)
      self.df_cpu = pd.concat([self.df_cpu, df_cpu_current], axis=0)
      self.df_mem = pd.concat([self.df_mem, df_mem_current], axis=0)
      self.df_cpu_history = pd.concat([self.df_cpu_history, df_cpu_current], axis=0, ignore_index=True)
      self.df_mem_history = pd.concat([self.df_mem_history, df_mem_current], axis=0, ignore_index=True)
      if self.csv_dir_name:
        self.df_cpu.to_csv(os.path.join(self.csv_dir_name, f'cpu_{self.csv_index:03d}.csv'), index=False)
        self.df_mem.to_csv(os.path.join(self.csv_dir_name, f'mem_{self.csv_index:03d}.csv'), index=False)
        if len(self.df_cpu) >= self.MAX_ROW_CSV:
          self.df_cpu = pd.DataFrame()
          self.df_mem = pd.DataFrame()
          self.csv_index += 1
      if len(self.df_cpu_history) >= self.MAX_NUM_HISTORY:
        self.df_cpu_history = self.df_cpu_history[1:]
        self.df_mem_history = self.df_mem_history[1:]

    self.df_cpu_history = self.sort_df_in_column(self.df_cpu_history)
    self.df_mem_history = self.sort_df_in_column(self.df_mem_history)

    return self.df_cpu_history, self.df_mem_history


  def reset_history(self):
    self.df_cpu_history = pd.DataFrame()
    self.df_mem_history = pd.DataFrame()


  @staticmethod
  def sort_df_in_column(df: pd.DataFrame):
    df = df.sort_values(by=len(df)-1, axis=1, ascending=False)
    return df


  @staticmethod
  def create_df_from_top(top_runner: TopRunner, lines: list[str], num_process: int):
    # now = datetime.datetime.now()
    now = int(time.time())
    for i, line in enumerate(lines):
      if 'PID' in line:
        lines = lines[i + 1:]
        break

    process_list = []
    cpu_list = []
    mem_list = []
    for i, line in enumerate(lines):
      if i >= num_process:
        break
      pid = line[top_runner.col_range_pid[0]:top_runner.col_range_pid[1]].strip()
      command = line[top_runner.col_range_command[0]:].strip()
      process_name = str(f'{command} ({pid})')
      process_list.append(process_name)
      cpu = float(line[top_runner.col_range_CPU[0]:top_runner.col_range_CPU[1]].strip())
      cpu_list.append(cpu)
      mem = float(line[top_runner.col_range_MEM[0]:top_runner.col_range_MEM[1]].strip())
      mem_list.append(mem)

    df_cpu_current = pd.DataFrame([[now] + cpu_list], columns=['datetime'] + process_list)
    df_mem_current = pd.DataFrame([[now] + mem_list], columns=['datetime'] + process_list)

    return df_cpu_current, df_mem_current



# Copyright 2023 iwatake2222
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging


def create_logger(name, level: int=logging.DEBUG, log_filename: str=None) -> logging.Logger:
  handler_format = logging.Formatter('[%(asctime)s][%(levelname)-7s][%(filename)s:%(lineno)s] %(message)s')
  # stream_handler = logging .StreamHandler()
  # stream_handler.setLevel(level)
  # stream_handler.setFormatter(handler_format)
  logger = logging.getLogger(name)
  logger.propagate = False
  logger.setLevel(level)
  # logger.addHandler(stream_handler)
  # if log_filename:
    # file_handler = logging.FileHandler(log_filename)
    # file_handler.setLevel(level)
    # file_handler.setFormatter(handler_format)
    # logger.addHandler(file_handler)
  return logger



# Copyright 2023 iwatake2222
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pandas as pd
import threading
import time
import dearpygui.dearpygui as dpg

from .data_container import DataContainer
from .top_runner import TopRunner
from .utility import create_logger


logger = create_logger(__name__, log_filename='rotop.log')

g_reset_history_df = False  # todo: add lock

COLOR_MAP = (
  # matplotlib.cm.tab20
  (31, 119, 180),
  (174, 199, 232),
  (256, 127, 14),
  (256, 187, 120),
  (44, 160, 44),
  (152, 223, 138),
  (214, 39, 40),
  (256, 152, 150),
  (148, 103, 189),
  (197, 176, 213),
  (140, 86, 75),
  (196, 156, 148),
  (227, 119, 194),
  (247, 182, 210),
  (127, 127, 127),
  (199, 199, 199),
  (188, 189, 34),
  (219, 219, 141),
  (23, 190, 207),
  (158, 218, 229),
  # matplotlib.cm.tab20b
  (57, 59, 121),
  (82, 84, 163),
  (107, 110, 207),
  (156, 158, 222),
  (99, 121, 57),
  (140, 162, 82),
  (181, 207, 107),
  (206, 219, 156),
  (140, 109, 49),
  (189, 158, 57),
  (231, 186, 82),
  (231, 203, 148),
  (132, 60, 57),
  (173, 73, 74),
  (214, 97, 107),
  (231, 150, 156),
  (123, 65, 115),
  (165, 81, 148),
  (206, 109, 189),
  (222, 158, 214),
)


class GuiView:
  def __init__(self):
    self.pause = False  # todo: add lock
    self.plot_is_cpu = True
    self.dpg_plot_axis_x_id = None
    self.dpg_plot_axis_y_id = None
    self.color_dict = {}
    self.theme_dict = {}


  def start_dpg(self):
    dpg.create_context()
    dpg.create_viewport(title='rotop')
    dpg.setup_dearpygui()

    with dpg.window(label='window', no_collapse=True, no_title_bar=True, no_move=True, no_resize=True) as self.dpg_window_id:
      with dpg.group(horizontal=True):
        self.dpg_button_cpumem = dpg.add_button(label='CPU/MEM', callback=self.cb_button_cpumem)
        self.dpg_button_reset = dpg.add_button(label='RESET', callback=self.cb_button_reset)
        self.dpg_button_pause = dpg.add_button(label='PAUSE', callback=self.cb_button_pause)
        dpg.add_text('Help(?)')
      with dpg.tooltip(dpg.last_item()):
          dpg.add_text('- CLick "Reset" to clear graph and history.\n- The data in the first half of the graph has been downsampled by 1/2 or 1/4.')
      with dpg.plot(label=self.get_plot_title(), use_local_time=True, no_title=True) as self.dpg_plot_id:
        self.dpg_plot_axis_x_id =  dpg.add_plot_axis(dpg.mvXAxis, label='datetime', time=True)
      self.dpg_text = dpg.add_text()

    dpg.set_viewport_resize_callback(self.cb_resize)
    self.cb_resize(None, [None, None, dpg.get_viewport_width(), dpg.get_viewport_height()])
    dpg.show_viewport()
    dpg.start_dearpygui()
    dpg.destroy_context()


  def get_plot_title(self):
    return 'CPU [%]' if self.plot_is_cpu else 'MEM [%]'


  def cb_button_cpumem(self, sender, app_data, user_data):
    self.plot_is_cpu = not self.plot_is_cpu
    dpg.set_item_label(self.dpg_plot_id, self.get_plot_title())


  def cb_button_reset(self, sender, app_data, user_data):
    global g_reset_history_df
    g_reset_history_df = True
    self.color_dict = {}
    self.theme_dict = {}


  def cb_button_pause(self, sender, app_data, user_data):
    self.pause = not self.pause


  def cb_resize(self, sender, app_data):
    window_width = app_data[2]
    window_height = app_data[3]
    dpg.set_item_width(self.dpg_window_id, window_width)
    dpg.set_item_height(self.dpg_window_id, window_height)
    dpg.set_item_width(self.dpg_plot_id, window_width)
    dpg.set_item_height(self.dpg_plot_id, window_height / 2)


  def update_gui(self, result_lines:list[str], df_cpu_history:pd.DataFrame, df_mem_history:pd.DataFrame):
    if self.pause:
      return
    if self.dpg_plot_axis_y_id:
      dpg.delete_item(self.dpg_plot_axis_y_id)
    self.dpg_plot_axis_y_id =  dpg.add_plot_axis(dpg.mvYAxis, label=self.get_plot_title(), lock_min=True, parent=self.dpg_plot_id)

    df = df_cpu_history if self.plot_is_cpu else df_mem_history
    col_x = df.columns[0]
    cols_y = df.columns[1:]

    x = df[col_x].to_list()
    mabiku_sep = None    # mabiku for speed
    mabiku_interval = 2
    if len(x) > DataContainer.MAX_NUM_HISTORY / 4:
      mabiku_interval = 4 if len(x) > DataContainer.MAX_NUM_HISTORY / 2 else 2
      mabiku_sep = int(len(x) / 4) * 2 # always even number
      x = x[0:mabiku_sep:mabiku_interval] + x[mabiku_sep:]

    for col_y in cols_y:
      y = df[col_y].to_list()
      if mabiku_sep:
        y = y[0:mabiku_sep:mabiku_interval] + y[mabiku_sep:]
      line_series = dpg.add_line_series(x, y, label=col_y[:min(40, len(col_y))].ljust(40), parent=self.dpg_plot_axis_y_id)
      theme = self.get_theme(col_y)
      dpg.bind_item_theme(line_series, theme)


    if  self.plot_is_cpu:
      dpg.add_line_series([x[0]], [110], label='', parent=self.dpg_plot_axis_y_id)  # dummy for ymax>=100
    dpg.add_plot_legend(parent=self.dpg_plot_id, outside=True, location=dpg.mvPlot_Location_NorthEast)
    dpg.fit_axis_data(self.dpg_plot_axis_x_id)
    dpg.fit_axis_data(self.dpg_plot_axis_y_id)

    dpg.set_value(self.dpg_text, '\n'.join(result_lines))


  def get_color(self, process_name)->tuple[int]:
    # return (0, 0, 0)
    if process_name in self.color_dict:
      return self.color_dict[process_name]
    else:
      color = COLOR_MAP[len(self.color_dict)%len(COLOR_MAP)]
      self.color_dict[process_name] = color
      return color

  def get_theme(self, process_name):
    if process_name in self.theme_dict:
      return self.theme_dict[process_name]
    else:
      with dpg.theme() as theme:
        with dpg.theme_component(dpg.mvLineSeries):
            dpg.add_theme_color(dpg.mvPlotCol_Line, self.get_color(process_name), category=dpg.mvThemeCat_Plots)
      self.theme_dict[process_name] = theme
      return theme


def gui_loop(view: GuiView):
  view.start_dpg()


def gui_main(args):
  global g_reset_history_df
  top_runner = TopRunner(args.filter, args.interval)
  data_container = DataContainer(args.csv)

  view = GuiView()
  gui_thread = threading.Thread(target=gui_loop, args=(view,))
  gui_thread.start()

  try:
    while True:
      result_lines, result_show_all_lines = top_runner.run(args.num_process, True)
      if result_show_all_lines is None:
        time.sleep(0.1)
        continue

      df_cpu_history, df_mem_history = data_container.run(top_runner, result_show_all_lines, args.num_process)
      df_cpu_history = df_cpu_history.iloc[:, :min(args.num_process, len(df_cpu_history.columns))]
      df_mem_history = df_mem_history.iloc[:, :min(args.num_process, len(df_mem_history.columns))]

      if gui_thread.is_alive():
        view.update_gui(result_lines, df_cpu_history, df_mem_history)

        if g_reset_history_df:
          data_container.reset_history()
          g_reset_history_df = False

      else:
        break
  except KeyboardInterrupt:
    dpg.stop_dearpygui()
    time.sleep(1)
    dpg.destroy_context()
    exit(0)

  gui_thread.join()



# Copyright 2023 iwatake2222
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import atexit
import pexpect
import re
import signal

from .utility import create_logger


logger = create_logger(__name__, log_filename='rotop.log')


class TopRunner:
  def __init__(self, filter, interval):
    self.child = pexpect.spawn(f'top -cb -d {interval} -o %CPU -w 512')
    self.filter_re = self.create_filter_re(filter)
    self.col_range_list_to_display = None
    self.col_range_pid = None
    self.col_range_CPU = None
    self.col_range_MEM = None
    self.col_range_command = None
    self.next_after = ''


  def __del__(self):
    signal.signal(signal.SIGINT, signal.SIG_IGN)  # ignore ctrl-c while closing
    self.child.close()


  def run(self, max_num_process, show_all=False):
    # get the result string of top command
    self.child.expect(r'top - .*load average:')
    before = self.child.before
    previous_after = self.next_after
    self.next_after = self.child.after
    if before == '' or previous_after == '' or self.next_after == '':
      return None, None
    top_str = (previous_after + before).decode('utf-8')
    orgial_lines = top_str.splitlines()

    result_lines = []
    result_show_all_lines = []
    row_process_info = 0

    # System Information
    for line in orgial_lines:
      result_lines.append(line)
      result_show_all_lines.append(line)
      if 'PID' in line:
        break

    # get layout information from process header line
    row_process_info = len(result_lines)
    process_header_org = result_lines[-1]
    self.analyze_cols(process_header_org, show_all)

    process_header = ''
    for range in self.col_range_list_to_display:
      process_header += process_header_org[range[0]:range[1]]
    result_lines[-1] = process_header

    # Process Information
    for line in orgial_lines[row_process_info:]:
      if self.col_range_command and self.col_range_command[0] > 0 and len(line) > self.col_range_command[0]:
        process_info_org = line[:self.col_range_command[0]]
        process_info = ''
        for range in self.col_range_list_to_display:
          process_info += process_info_org[range[0]:range[1]]
        command_str = line[self.col_range_command[0]:]
        if not self.filter_re.match(command_str):
          continue
        command_str = self.parse_command_str(command_str)

        line = process_info + command_str
        show_all_line = process_info_org + command_str

        result_lines.append(line)
        result_show_all_lines.append(show_all_line)
        if len(result_lines) >= row_process_info + max_num_process:
          break

    return result_lines, result_show_all_lines


  def analyze_cols(self, process_header: str, show_all: bool):
    if self.col_range_command is None or self.col_range_command[0] == -1:
      self.col_range_list_to_display = self.get_col_range_list_to_display(process_header, show_all)
      self.col_range_pid = TopRunner.get_col_range_PID(process_header)
      self.col_range_CPU = TopRunner.get_col_range_CPU(process_header)
      self.col_range_MEM = TopRunner.get_col_range_MEM(process_header)
      self.col_range_command = TopRunner.get_col_range_command(process_header)
    return



  @staticmethod
  def create_filter_re(filter_str):
    if '.*' not in filter_str:
      filter_str = '.*' + filter_str + '.*'
    filter_re = re.compile(filter_str)
    return filter_re


  @staticmethod
  def get_row_start_list(lines: list[str])->list[int]:
    row_list = []
    for i, line in enumerate(lines):
      if 'top' in line and 'load average' in line:
        row_list.append(i)
    return row_list


  @staticmethod
  def get_col_range_command(process_info_header_line: str):
    start_col = process_info_header_line.find('COMMAND')
    end_col = len(process_info_header_line) - 1
    return (start_col, end_col)


  @staticmethod
  def get_col_range_PID(process_info_header_line: str):
    start_col = 0
    end_col = process_info_header_line.find('PID') + len('PID')
    return (start_col, end_col)


  @staticmethod
  def get_col_range_CPU(process_info_header_line: str):
    start_col = process_info_header_line.find('SHR S') + len('SHR S')
    end_col = process_info_header_line.find('%CPU') + len('%CPU')
    return (start_col, end_col)


  @staticmethod
  def get_col_range_MEM(process_info_header_line: str):
    start_col = process_info_header_line.find('%CPU') + len('%CPU')
    end_col = process_info_header_line.find('%MEM') + len('%MEM')
    return (start_col, end_col)


  @staticmethod
  def get_col_range_list_to_display(process_info_header_line: str, show_all=False):
    range_list = []

    if show_all:
      range_list.append((0, len(process_info_header_line)))
    else:
      start_col = 0
      end_col = process_info_header_line.find('PID') + len('PID')
      range_list.append((start_col, end_col))

      start_col = process_info_header_line.find('NI') + len('NI')
      end_col = process_info_header_line.find('%MEM') + len('%MEM')
      range_list.append((start_col, end_col))

      start_col = process_info_header_line.find('COMMAND') - 1
      end_col = len(process_info_header_line)
      range_list.append((start_col, end_col))

    return range_list


  @staticmethod
  def parse_component_container_command(command):
    cmd = command.split()[0].split('/')[-1]
    idx_node = command.find('__node')
    if idx_node > 0:
      node = command[idx_node:].split()[0].split('=')[-1]
      cmd = node
    idx_ns = command.find('__ns')
    if idx_ns > 0:
      ns = command[idx_ns:].split()[0].split('=')[-1]
      # cmd = cmd + ', ' + node + ', ' + ns
      cmd += ', ' + ns
    return cmd


  @staticmethod
  def parse_python_command(command):
    cmd_list = command.split()
    cmd = cmd_list[0].split('/')[-1]
    if len(cmd_list) > 1:
      if cmd_list[1][0] == '-':
          python_file = cmd_list[-1]
      else:
          python_file = cmd_list[1]
      python_file = python_file.split('/')[-1]

      ros2_option = ''
      if 'ros2' == python_file:
          ros2_option = ' '.join(cmd_list[2:5])

      cmd = cmd + ' ' + python_file + ' ' + ros2_option
    return cmd


  @staticmethod
  def parse_command_str(command):
    param_for_ros2 = ['__node', '__ns']
    if '[' == command[0]:
        # kernel process
        command = command
    elif any(item in command for item in param_for_ros2):
        command = TopRunner.parse_component_container_command(command)
    elif 'python' in command:
        command = TopRunner.parse_python_command(command)
    else:
        # normal process
        command = command.split()[0].split('/')[-1]
    return command


# Copyright 2023 iwatake2222
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import curses
import time

from .data_container import DataContainer
from .top_runner import TopRunner
from .gui_main import gui_main
from .utility import create_logger


logger = create_logger(__name__, log_filename='rotop.log')


def main_curses(stdscr, args):
  curses.use_default_colors()
  # curses.init_color(0, 0, 0, 0)
  curses.curs_set(0)
  stdscr.timeout(100)

  top_runner = TopRunner(args.filter, args.interval)
  data_container = DataContainer(args.csv)

  try:
    while True:
      max_y, max_x = stdscr.getmaxyx()
      result_lines, result_show_all_lines = top_runner.run(max(max_y, args.num_process), max_x>160)
      if result_show_all_lines is None:
        time.sleep(0.1)
        continue

      _ = data_container.run(top_runner, result_show_all_lines, args.num_process)

      stdscr.clear()
      for i, line in enumerate(result_lines):
        if i >= max_y - 1:
          break
        stdscr.addstr(i, 0, line[:max_x])

      stdscr.refresh()
      key = stdscr.getch()
      if key == ord('q'):
        break
  except KeyboardInterrupt:
    exit(0)


def parse_args():
  parser = argparse.ArgumentParser(
    description='rotop: top for ROS 2')
  parser.add_argument('--interval', type=float, default=2)
  parser.add_argument('--filter', type=str, default='.*')
  parser.add_argument('--csv', action='store_true', default=False)
  parser.add_argument('--gui', action='store_true', default=False)
  parser.add_argument('--num_process', type=int, default=30)
  args = parser.parse_args()

  logger.debug(f'filter: {args.filter}')
  logger.debug(f'csv: {args.csv}')
  logger.debug(f'gui: {args.gui}')
  logger.debug(f'num_process: {args.num_process}')

  return args


def main():
  args = parse_args()
  if args.gui:
    gui_main(args)
  else:
    curses.wrapper(main_curses, args)
